// Copyright (c) 2023 mStar
//
// Licensed under the EUPL, Version 1.2
//
// You may not use this work except in compliance with the Licence.
// You should have received a copy of the Licence along with this work. If not, see:
// <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
// See the Licence for the specific language governing permissions and limitations under the Licence.
//

use anyhow::{anyhow, bail, Error};
use clap::Parser;
use serde_derive::Deserialize;
use std::{fs, path::PathBuf};
use sysinfo::System;

const MEGABYTE: u64 = 1049000;
/// Assumed memory consumption per worker
/// Rather overshoot this than undershoot
const ASSUMED_MEM_PER_WORKER: u64 = 2 * MEGABYTE;
const CPUS_RESERVED_FOR_SYSTEM: u32 = 2;

#[derive(Debug)]
pub struct Config {
    /// Path to the db for diesel to use
    pub database_path: String,
    /// Maximum number of workers to use
    /// A value of 0 makes stuff single-threaded
    pub max_workers: u32,
    /// Minimum number of workers to use
    pub min_workers: u32,
}

#[derive(Debug, Deserialize)]
struct ConfigRootToml {
    config: Option<ConfigToml>,
}

#[derive(Debug, Deserialize)]
struct ConfigToml {
    /// Path to the db for diesel to use
    database_path: Option<String>,
    /// Maximum number of workers to use
    /// A value of 0 makes stuff single-threaded
    max_workers: Option<u32>,
    /// Minimum number of workers to use
    min_workers: Option<u32>,
}

/// An ActivityPub enabled social media server
#[derive(Debug, Parser)]
#[command(author, version, about, long_about = None)]
pub struct CLIArguments {
    /// Path to the config file
    #[arg(default_value = "config.toml")]
    pub config: PathBuf,
    /// Maximum amount of workers to use.
    /// If not set, the max is calculated depending on system resources.
    /// A value of 0 will cause the server to be singlethreaded, potentially reducing the performance and responsiveness by a lot
    #[arg(long)]
    pub max_workers: Option<u32>,
    /// Minimum amount of workers to keep active at all times
    /// Can potentially increase reaction speed at the cost of resource usage
    #[arg(long)]
    pub min_workers: Option<u32>,

    #[arg(long)]
    pub db_path: Option<String>,
}

pub fn read_from_env() -> Result<Config, Error> {
    let args = CLIArguments::parse();

    if !args.config.is_file() {
        bail!("Config path is not a file");
    };
    let ext = args.config.extension().ok_or(anyhow!(
        "Type of config file couldn't be determined. No extension"
    ))?;
    if ext != "toml" {
        bail!("Config file is not a toml file");
    }

    let contents = match fs::read_to_string(&args.config) {
        Ok(c) => c,
        Err(err) => bail!("Couldn't read config file. Error: {err}"),
    };

    let unsanitized: ConfigRootToml = match toml::from_str(&contents) {
        Ok(d) => d,
        Err(err) => bail!("Failed to parse config file. Error: {err}"),
    };
    let conf = unsanitized.into_config(&args);
    conf.sanity_check()?;
    Ok(conf)
}

fn calc_max_workers_from_system() -> u32 {
    let mut sys = System::new_all();
    sys.refresh_all();
    let available_ram = sys.available_memory() - sys.used_memory();
    let available_cpus = (sys.cpus().len() as u32) - CPUS_RESERVED_FOR_SYSTEM;
    let max_workers_ram = (available_ram / ASSUMED_MEM_PER_WORKER) as u32;
    // Assume one worker per available CPU
    // The lower number of potential workers is what we'll use
    if available_cpus > max_workers_ram {
        max_workers_ram
    } else {
        available_cpus
    }
}

impl ConfigRootToml {
    fn into_config(&self, cli: &CLIArguments) -> Config {
        match &self.config {
            Some(config) => {
                let db_path = match &cli.db_path {
                    Some(path) => path.clone(),
                    None => match &config.database_path {
                        Some(path) => path.clone(),
                        None => "config.toml".into(),
                    },
                };
                let min_workers = match cli.min_workers {
                    Some(num) => num,
                    None => match config.min_workers {
                        Some(num) => num,
                        None => 1,
                    },
                };
                let max_workers = match cli.max_workers {
                    Some(num) => num,
                    None => match config.max_workers {
                        Some(num) => num,
                        None => calc_max_workers_from_system(),
                    },
                };
                Config {
                    database_path: db_path,
                    min_workers,
                    max_workers,
                }
            }
            None => {
                let db_path = match &cli.db_path {
                    Some(path) => path.clone(),
                    None => "config.toml".to_string(),
                };
                let min_workers = match cli.min_workers {
                    Some(num) => num,
                    None => 1,
                };
                let max_workers = match cli.max_workers {
                    Some(num) => num,
                    None => calc_max_workers_from_system(),
                };
                Config {
                    database_path: db_path.clone(),
                    min_workers,
                    max_workers,
                }
            }
        }
    }
}

impl Config {
    fn sanity_check(&self) -> Result<(), Error> {
        if self.min_workers > self.max_workers {
            bail!("Minimum amount of workers ({}) can't be greater than maximum amount of workers ({})", self.min_workers, self.max_workers)
        };
        Ok(())
    }
}