// Copyright (c) 2023 mStar
// 
// Licensed under the EUPL, Version 1.2
// 
// You may not use this work except in compliance with the Licence.
// You should have received a copy of the Licence along with this work. If not, see:
// <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
// See the Licence for the specific language governing permissions and limitations under the Licence.
// 

// TODO: Once diesel-async supports sqlite (see https://github.com/weiznich/diesel_async/discussions/95), convert to async
pub mod pg_schemas;
pub mod pg_person;
pub mod pg_post;

use anyhow::{Error, bail};
use diesel::{pg::Pg, r2d2::{Pool, ConnectionManager, self}};
use diesel::PgConnection;
use diesel_migrations::{EmbeddedMigrations, embed_migrations, MigrationHarness};

pub const POSTGRES_MIGRATIONS: EmbeddedMigrations = embed_migrations!("migrations");


#[derive(Clone)]
pub struct Storage {
    pub pool: Pool<ConnectionManager<PgConnection>>,
}

/// Try and connect to a database at the given url
/// The url can be a path to an sqlite db file or an url to a postgres db
pub fn establish_connection(db_url: &str) -> Result<Storage, Error> {
    let manager = ConnectionManager::<PgConnection>::new(db_url);
    let pool = match r2d2::Pool::builder().build(manager) {
        Ok(a) => a,
        Err(err) => bail!("Error while creating the connection pool. Error {}", err)
    };
    match pool.get() {
        Ok(mut conn) => {run_migrations_postgres(&mut conn)?;}
        Err(e) => bail!(e)
    }
    Ok(Storage{
        pool
    })
}

fn run_migrations_postgres(conn: &mut impl MigrationHarness<Pg>) -> Result<(), Error> {
    match conn.run_pending_migrations(POSTGRES_MIGRATIONS) {
        Ok(_) => Ok(()),
        Err(e) => bail!("Migrations failed. Error {}", e)
    }
}
