use activitypub_federation::{traits::Object, config::Data, fetch::object_id::ObjectId, protocol::verification::verify_domains_match};
use anyhow::{anyhow, Error};
use chrono::{Utc, NaiveDateTime, DateTime};
use diesel::prelude::*;
use url::Url;

use crate::objects::person::Person;

use super::{Storage, pg_schemas};


#[derive(Queryable, Selectable, Debug, Clone)]
#[diesel(table_name = crate::storage::pg_schemas::people)]
#[diesel(check_for_backend(diesel::pg::Pg))]
pub struct PersonPG {
    pub ap_id: String,
    pub name: String,
    pub instance: Option<String>, // Also url
    pub mail: Option<String>,
    pub pw_hash: Option<String>,
    pub inbox: String, // Is an url but diesel no likey
    pub public_key: String,
    pub private_key: Option<String>,
    pub last_refreshed_at: NaiveDateTime, 
    pub local: bool,
    pub followers: Vec<Option<String>> // Strings are URLs. This is stupid. I wish I could simply use Vec<Url> instead
}

#[async_trait::async_trait]
impl Object for PersonPG {
    type DataType = Storage;
    type Kind = Person;
    type Error = Error;

    fn last_refreshed_at(&self) -> Option<DateTime<Utc>>  {
        Some(self.last_refreshed_at.and_utc())
    }

    async fn read_from_id(
        object_id: Url,
        data: &Data<Self::DataType>,
    ) -> Result<Option<Self>, Self::Error> {
        use pg_schemas::people::dsl::*;
        let pool = data.pool.clone();
        let mut conn = pool.get()?;
        let results: Vec<PersonPG> = people
            .filter(ap_id.eq(object_id.as_str()))
            .limit(1)
            .select(PersonPG::as_select())
            .load(&mut conn)
            .expect(format!("Error loading person from id {}", object_id).as_str());
        if results.len() == 1 {
            Ok(Some(results[0].clone()))
        } else {
            Ok(None)
        }
    }

    async fn into_json(self, _data: &Data<Self::DataType>) -> Result<Self::Kind, Self::Error> {
        Ok(Person {
            kind: Default::default(),
            preferred_username: self.name.clone(),
            id: ObjectId::parse(&self.ap_id)?,
            inbox: Url::parse(&self.inbox)?,
            public_key: serde_json::from_str(&self.public_key).unwrap(),
        })
    }

    async fn verify(
        json: &Self::Kind,
        expected_domain: &Url,
        _data: &Data<Self::DataType>,
    ) -> Result<(), Self::Error> {
        Ok(verify_domains_match(&json.id.inner(), &expected_domain)?)
        
    }

    async fn from_json(
        json: Self::Kind,
        _data: &Data<Self::DataType>,
    ) -> Result<Self, Self::Error> {
        Ok(
            PersonPG { 
                ap_id: json.id.inner().to_string(),
                name: json.preferred_username,
                instance: Some(json.inbox.authority().to_string()),
                mail: None,
                pw_hash: None,
                inbox: json.inbox.to_string(),
                public_key: serde_json::to_string(&json.public_key).unwrap(),
                private_key: None,
                last_refreshed_at: Utc::now().naive_utc(),
                local: false,
                followers: vec![]
            }
        )
    }
}